//---------------------------------------------------------------------------
#include "AJBLisp.h"

const int EVALLIMIT = 3;
//---------------------------------------------------------------------------

LISP Eval(LISP c)
{
  if(UserBreak)
  {
    UserBreak = false;
    LispError(23);  // User break
  }
  if(!c) return c;
  // Symbols evaluate to their value
  if(IsSymbol(c)) return c->value;
  // Other atoms evaluate to themselves
  if(IsAtom(c)) return c;
  // Then we must have a list with the function in the car and
  // args (if any) in the cdr
  PP* ppsave = PPLIST;
  volatile LISP func = NIL;
  PP pfunc;
  protect(&func, &pfunc);
  // Build a frame
  NewFrame(c);
  func = c->car;
  // Try to find a function
  for(int i = 0; i < EVALLIMIT; i++)
  {
    if(!func)
      LispError(11);      // Illegal function
    if(IsSubr(func))
    {
      PPLIST = ppsave;
      return EvalSubr(func);
    }
    if(IsFSubr(func))
    {
      PPLIST = ppsave;
      return EvalFSubr(func);
    }
    if(func->car == LAMBDA)
    {
      if(!func->cdr)
        LispError(11);    // Illegal function

      if(!(func->cdr->car) || IsAtom(func->cdr->car))
      {
        PPLIST = ppsave;
        return EvalFExpr(func);
      }
      else
      {
        PPLIST = ppsave;
        return EvalExpr(func);
      }
    }
    func = Eval(func);    // Evaluate to see if we can find a function
  }
  //DumpHeap();
  LispError(11);          // Illegal function
  return NIL;
}

LISP EvalSubr(LISP f)
{
  LISP result;
  // Get the execution address before everything moves
  SUBR exec = (SUBR)(f->code);
  // Evaluate list of args
  EvList(base->fun->cdr);
  result = exec(base->nargs, base->args);
  PopFrame();
  return result;
}

LISP EvalFSubr(LISP f)
{
  LISP result;
  // Push list of args
  Push(base->fun->cdr);
  base->nargs++;
  result = ((FSUBR)(f->code))(base->args[0]);
  PopFrame();
  return result;
}

LISP EvalExpr(LISP f)
{
  PP* ppsave = PPLIST;
  volatile LISP body = NIL;
  PP pbody;
  protect(&body, &pbody);
  volatile LISP func = f;
  PP pfunc;
  protect(&func, &pfunc);
  base->save = true;               // This is a save frame
  EvList(base->fun->cdr);          // Evaluate args into frame
  Bind(func->cdr->car);            // Bind formal args to values in frame
  LISP result = NIL;               // Init. result for function with null body
  body = func->cdr->cdr;
  while(body)                      // Implicit PROGN in body code
  {
    result = Eval(body->car);      // Evaluate body code
    body = body->cdr;
  }
  PopFrame();                      // Restore environment
  PPLIST = ppsave;
  return result;
}

LISP EvalFExpr(LISP f)
{
  PP* ppsave = PPLIST;
  volatile LISP body = NIL; 
  PP pbody;
  protect(&body, &pbody);
  LISP farg = f->cdr->car;
  // Functions with no arguments are treated as FSUBRs because it is more
  // efficient.  Hence the following condition...
  if(farg)
  {
    if(!IsSymbol(farg))          // Formal arg must be a symbol
      LispError(12);             // Illegal formal argument
    base->save = true;           // This is a save frame
    base->nargs = 1;             // ...with 1 argument
    Push(farg);                  // Save argument
    Push(farg->value);
    Push(farg->prop);
    farg->value = base->fun->cdr;// Initialise formal argument
    farg->prop = NULL;
  }
  LISP result = NIL;             // Init. result for function with null body
  body = f->cdr->cdr;
  while(body)                    // Implicit PROGN in body code
  {
    result = Eval(body->car);    // Evaluate body code
    body = body->cdr;
  }
  PopFrame();                    // Restore environment
  PPLIST = ppsave;
  return result;
}

//---------------------------------------------------------------------------



