#include "AJBLisp.h"
#include <excpt.h>
#include <float.h>
//---------------------------------------------------------------------------

bool UserBreak;   // If true the Eval will be interrupted

void REPL()
{
  Error aritherror;
  // Main interpreter loop
  while(true)
  {
    aritherror.err = 0;
    try      // __except (Windows SEH)
    {
      try    // catch
      {
        fputs("> ", stdout);
        LISP cell = read(stdin);
        // exit if 'exit' entered
        if(IsSymbol(cell) && !strncmp(cell->name, "EXIT", 4))
        {
          FreeHeap();
          FreeStack();
          exit(0);
        }
        PPLIST = 0;
        UserBreak = false;
        cell = Eval(cell);
        fputs("= ", stdout);
        print(cell, stdout, false);
        putchar('\n');
      }
      catch(Error e)
      {
        ReportError(e);
      }
    }
    __except(FilterException(GetExceptionCode()))
    {
      aritherror = OSException(GetExceptionCode());
    }
    // Report errors trapped by __except
    if(aritherror.err)
      ReportError(aritherror);
  }
}

//---------------------------------------------------------------------------
void LoadFile(char *filename)
{
  // Load a lisp source file
  Error aritherror;
  aritherror.err = 0;
  FILE* f = fopen(filename, "r");
  if(f)
  {
    try      // __except (Windows SEH)
    {
      try    // catch
      {
        while(!feof(f))
        {
          LISP in = read(f);
          UserBreak = false;
          Eval(in);
        }
      }
      catch(Error e)
      {
        fclose(f);
        ReportError(e);
        return;
      }
    }
    __except(FilterException(GetExceptionCode()))
    {
      fclose(f);
      aritherror = OSException(GetExceptionCode());
    }
    // Report errors trapped by __except
    if(aritherror.err)
      ReportError(aritherror);
    else
      fclose(f);
  }
}

//---------------------------------------------------------------------------
void ReportError(Error e)
{
  char nummsg[20];

  if(!e.msg) // If no user-supplied message then supply the default
  {
    switch(e.err)
    {
      // Fatal error messages
      case  0: break;    // Silent exit
      case -1: e.msg = "Failed to allocate heap"; break;
      case -2: e.msg = "Failed to allocate stack"; break;
      case -3: e.msg = "Out of heap space"; break;
      case -4: e.msg = "Out of stack space"; break;
      case -5: e.msg = "Corrupted lisp stack"; break;
      case -6: e.msg = "Failed to allocate string heap"; break;
      // Non-fatal error messages
      case  1: e.msg = "Programmed error"; break;
      case  2: e.msg = "Incorrect number of arguments to function"; break;
      case  3: e.msg = "List argument expected"; break;
      case  4: e.msg = "Numeric argument expected"; break;
      case  5: e.msg = "Integer argument expected"; break;
      case  6: e.msg = "Real argument expected"; break;
      case  7: e.msg = "Symbol argument expected"; break;
      case  8: e.msg = "String argument expected"; break;
      case  9: e.msg = "File argument expected"; break;
      case 10: e.msg = "Atomic argument expected"; break;
      case 11: e.msg = "Illegal function"; break;
      case 12: e.msg = "Illegal formal argument"; break;
      case 13: e.msg = "Cannot change value of special symbol"; break;
      case 14: e.msg = "Badly formed association list"; break;
      case 15: e.msg = "Badly formed property list"; break;
      case 16: e.msg = "Aborted null loop"; break;
      case 17: e.msg = "Out of string space"; break;
      case 18: e.msg = "Symbol name too long"; break;
      case 19: e.msg = "Too many arguments for stack frame"; break;
      case 20: e.msg = "Cannot interpret string as float"; break;
      case 21: e.msg = "Cannot interpret string as integer"; break;
      case 22: e.msg = "Symbol or numeric argument expected"; break;
      case 23: e.msg = "User interrupt"; break;
      case 24: e.msg = "Badly formed binding in LET"; break;
      case 25: e.msg = "Illegal ASCII value"; break;
      case 26: e.msg = "String too long"; break;
      case 27: e.msg = "Symbol, numeric or list argument expected"; break;
      // Maths errors
      // Maths errors raised within the interpreter will have more
      // helpful error messages which specify the function that
      // failed.  These messages are provided for user.
      case 30: e.msg = "Maths - domain error"; break;
      case 31: e.msg = "Maths - result singular"; break;
      case 32: e.msg = "Maths - result overflow"; break;
      case 33: e.msg = "Maths - result underflow"; break;
      case 34: e.msg = "Maths - result inaccurate"; break;
      case 35: e.msg = "Maths - invalid operation"; break;
      // I/O errors
      case 40: e.msg = "EOF after reading escape character"; break;
      case 41: e.msg = "EOF inside string"; break;
      case 42: e.msg = "Illegal character inside string"; break;
      case 43: e.msg = "Unexpected close parenthesis"; break;
      case 44: e.msg = "EOF inside list"; break;
      case 45: e.msg = "Missing close parenthesis"; break;
      case 46: e.msg = "List argument to patom"; break;
      case 47: e.msg = "Input token exceeds buffer size (241 characters)"; break;
      case 48: e.msg = "Illegal read macro"; break;
      // Debug
      //case 99: e.msg = "Debug error"; break;
      // Unrecognised
      default: sprintf(nummsg, "Error number %d", e.err);
               e.msg = nummsg;
               break;
    }
  }
  if(e.err > 0)  // Non fatal errors
  {
    printf("\nError: %s\n", e.msg);
    ClearStack();
    fflush(stdin);
    return;
  }
  // Going down!
  if(e.err < 0) printf("\nFatal: %s\n", e.msg);
  else if(e.msg) printf("%s\n", e.msg);
  FreeHeap();
  FreeStack();
  exit(-e.err);
}

// What follows is ugly, non-portable stuff
// If you read it just remember I warned you....

// Handle errors raised by the maths library
#ifdef __BORLANDC__
// Disregard the C++ Builder help files
// This is the only definition of _matherr() which works
int cdecl std::_matherr(struct std::math_exception *e)
#else
// The rest-of-the-world's version...
int _matherr(struct _exception *e)
#endif
{
  char  msgbuf[40];
  char* msg;
  Error le;
  le.msg = msgbuf;
  switch(e->type)
  {
    case DOMAIN:
      msg = "Domain error in ";
      le.err = 30;
    break;
    case SING:
      msg = "Singular result in ";
      le.err = 31;
    break;
    case OVERFLOW:
      msg = "Overflow in ";
      le.err = 32;
    break;
    case UNDERFLOW:
      msg = "Underflow in ";
      le.err = 33;
    break;
    case TLOSS:
      msg = "Total loss of accuracy in ";
      le.err = 34;
    break;
  }
  // C++ uses different names for some functions
  char* name = e->name;
  if(!strcmp(name, "log"))   name = "ln";
  if(!strcmp(name, "log10")) name = "log";
  if(!strcmp(name, "atan2")) name = "atan";
  strcpy(msgbuf, msg);
  strcat(msgbuf, name);
  throw le;
}

// Indicate which execeptions are handled
int FilterException(int code)
{
  switch(code)
  {
    case EXCEPTION_FLT_DIVIDE_BY_ZERO:
    case EXCEPTION_FLT_INEXACT_RESULT:
    case EXCEPTION_FLT_INVALID_OPERATION:
    case EXCEPTION_FLT_OVERFLOW:
    case EXCEPTION_FLT_UNDERFLOW:
    case EXCEPTION_INT_DIVIDE_BY_ZERO:
    case EXCEPTION_INT_OVERFLOW:
      return EXCEPTION_EXECUTE_HANDLER;    // We handle these
    default:
      return EXCEPTION_CONTINUE_SEARCH;    // ...but no others
  }
}

// A structured exception has been delivered by Windows
// Translate it into a lisp error
Error OSException(int code)
{
  Error aritherror;
  switch(code)
  {
    case EXCEPTION_FLT_DIVIDE_BY_ZERO:
      aritherror.msg = "Floating point divide by zero";
      aritherror.err = 31;
    break;
    case EXCEPTION_FLT_INEXACT_RESULT:
      aritherror.msg = "Floating point total loss of accuracy";
      aritherror.err = 34;
    break;
    case EXCEPTION_FLT_INVALID_OPERATION:
      aritherror.msg = "Invalid floating point operation";
      aritherror.err = 35;
    break;
    case EXCEPTION_FLT_OVERFLOW:
      aritherror.msg = "Floating point overflow";
      aritherror.err = 32;
    break;
    case EXCEPTION_FLT_UNDERFLOW:
      aritherror.msg = "Floating point underflow";
      aritherror.err = 33;
    break;
    case EXCEPTION_INT_DIVIDE_BY_ZERO:
      aritherror.msg = "Integer divide by zero";
      aritherror.err = 31;
    break;
    case EXCEPTION_INT_OVERFLOW:
      aritherror.msg = "Integer overflow";
      aritherror.err = 32;
    break;
  }
  _fpreset();
  return aritherror;
}

