; defun is declared using setq for obvious reasons
; Note: the odd variable name, '', is used to avoid the funarg problem
;       Anyone declaring a function with this name deserves trouble!!
(setq defun
  '(lambda \''
     (set (car \'')
          (cons 'lambda (cdr \'')) )
     (car \'') ))

; Get the last top-level element of a list
(defun last (l)
  (if (null (cdr l))
      (car l)
      (last (cdr l)) ))

; Standard list append operation
(defun append (a b)
  (if (not a)
      b
      (cons (car a) 
            (append (cdr a) b) )))

; Insertion sort
(defun sort (l)
  (if (not l)
      nil
      (insert (car l) (sort (cdr l))) ))

; Insert a number into a sorted list
(defun insert (x l)
  (if (not l)
      (list x)
      (if (< x (car l))
          (cons x l)
          (cons (car l)(insert x (cdr l))) )))


; Map a function onto a list
(defun mapc(~x ~l)
  (if (null ~l)
      nil
      (cons (~x (car ~l)) (mapc ~x (cdr ~l)))))

; Substitute a for b in a list c
(defun subst (a b c)
  (cond
    ((eq b c) a)
    ((atom c) c)
    (t (cons (subst a b (car c))
             (subst a b (cdr c)) ))))

; Reverse a list
(defun reverse(l)
  (let ((W))
    (loop
      (while l w)
      (setq w (cons (car l) w))
      (setq l (cdr l)) )))

; Delete top level atoms from a list
(defun delete (x l)
  (cond
    ((null l) nil)
    ((eq (car l) x) (delete x (cdr l)))
    (t (cons (car l) (delete x (cdr l)))) ))

; Print newline
(defun terpri () (print))

; Read until valid input received
; Ignores invalid input following valid input
; e.g entering: 1 2 3) will read 1
(defun saferead((prompt . ">> ") (msg . "Error in input - Please re-enter"))
  (let ((x))
    (loop
      (princ prompt)
      (flush)
      (setq x (errorset (read)))
      (until (listp x) (flush) (car x))
      (printc msg) )))
    
; A very messy apply
; There does not seem to be a simple way to implement apply
; in the interpreter due to the fact that eval is somewhat monolithic
; (There is no apply phase in eval)
(defun apply(f a)
  (eval (cons f (mapc '(lambda (x)(list 'quote x)) a))) )
  
; Trace function execution
; Don't trace functions used in trace itself (for obvious reasons!)
(defun trace fn
  (setq fn (car fn))
  (put fn 'olddef (eval fn))
  (set fn (subst fn 'fn
                 '(lambda *x
                    (setq *x (mapc eval *x))
                    (printc 'fn space *x cr)
                    (setq *x (apply (get 'fn 'olddef) *x))
                    (printc 'fn '= *x cr)
                    *x)))
  fn)

(defun untrace fn
  (setq fn (car fn))
  (set fn (get fn 'olddef))
  (remprop fn 'olddef)
  (list fn 'untraced))


